/*
 * Copyright (C) 1995, 1996  Warren Jasper
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 */

/***************************************************************************
 *
 * acl8216.c
 *
 ***************************************************************************/

#define MODULE

/* retrieve the CONFIG_* macros */
#include <linux/autoconf.h>
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#define MODVERSIONS
#endif

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/version.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/mm.h>
#include <linux/timer.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/uaccess.h>
//#include "dda06-16.h"
#include "acl8216.h"

/***************************************************************************
 *
 * Prototype of pubic and private functions.
 *
 ***************************************************************************/

static int adc_read(struct file *filePtr, char *buf, size_t size, loff_t *offp);
static int adc_write(struct file *filePtr, const char *buf, size_t size, loff_t *offp);
static int adc_open(struct inode *iNode, struct file *filePtr);
static int adc_close(struct inode *iNode, struct file *filePtr);
static int adc_ioctl(struct inode *iNode, struct file *filePtr, unsigned int cmd, LONG arg);
static int adc_find( WORD base_reg );
int init_module(void);
void cleanup_module(void) ;

/***************************************************************************
 *
 * Global data.
 *
 ***************************************************************************/

static WORD base_reg =  PORT_MIN;              /* base register address         */
//static BYTE dio_reg = 0x80;                    /* status register 1             */
static int MajorNumber = DEFAULT_MAJOR_DEV;    /* Major number compiled in      */
static ChanRec Chan[AD_CHANNELS];    /* Channel specific information  */

#if LINUX_VERSION_CODE < 0x020000
char kernel_version[] = UTS_RELEASE;           /* Required by module            */
#endif

/***************************************************************************
 *
 * read() service function
 *
 ***************************************************************************/

static int adc_read(struct file *filePtr, char *buf, size_t size, loff_t *offp)
{
  BYTE  lsb,msb;
  WORD	value;
  int i;
  int   minor = MINOR(filePtr->f_dentry->d_inode->i_rdev);

  if ( minor >= 0 && minor < AD_CHANNELS ) {

	/* Set Range */
	outb_p( 0x00, base_reg + 9);

	/* Channel Multiplexer: Channel minor */
	if( minor < 8 ) {
		outb_p( 0x10 + minor, base_reg + 10);
	} else {
		outb_p( 0x28 + minor, base_reg + 10);
	}

//	outb_p( 0x10, base_reg + 10);
	current->state=TASK_INTERRUPTIBLE;
	schedule_timeout(1);

	outb_p( 0x01, base_reg + 11); /* Set Software Trigger Mode */

	outb_p( 0xff, base_reg + 12); /* Software Trigger */


//	current->state=TASK_INTERRUPTIBLE;
//	schedule_timeout(10*HZ/1000);

	for(i=0;i<100000;i++)
		if( !((lsb=inb_p(base_reg + 8)) &0x20) ) break;

//	if(i==100000) return 0;
//	printk("%d\n",lsb);

	lsb=inb_p(base_reg + 4);
	msb=inb_p(base_reg + 5);
	value=(WORD)msb*256+lsb;
	copy_to_user(buf,&value,2);
	return 2;
  }

  printk("dac_read: Incorrect minor number (%d).\n", minor);
  return -1;
}

/***************************************************************************
 *
 * write() service function
 *
 ***************************************************************************/

static int adc_write(struct file *filePtr, const char *buf, size_t size, loff_t *offp)
{
#if 0
#error Later
  int  minor = MINOR(iNode->i_rdev);
  BYTE value;

  if ( minor >= 0 && minor < AD_CHANNELS ) {   /* write out D/A value */
      Chan[minor].value = get_fs_word(buf);
      value = get_fs_byte(buf);
      outb( value, Chan[minor].addr);
      value = get_fs_byte(buf+1);
      outb( value, Chan[minor].addr+1);
      return 1;                                /* return number of WORDS written */
  }

  /* check to see if reading a value from the DIO */
  if ( minor >= AD_CHANNELS && minor < AD_CHANNELS + DIO_PORTS ) {
    value = get_fs_byte(buf);
    outb_p( value, Chan[minor].addr);
    
#ifdef DEBUG
    printk("%s: DIO Port %#x set to %#x\n", ADAPTER_ID, minor-AD_CHANNELS, value);
#endif
    return 1;
  }
#endif
  return -1;
}


/***************************************************************************
 *
 * open() service handler
 *
 ***************************************************************************/

static int adc_open(struct inode *iNode, struct file *filePtr)
{
  int minor = MINOR(iNode->i_rdev);

  /* 
     check if device is already open: only one process may read from a
     port at a time.  There is still the possibility of two processes 
     reading from two different channels messing things up. However,
     the overhead to check for this may not be worth it.
  */
  if ( !(minor >= 0 && minor < AD_CHANNELS) ) {
  	return -EINVAL;
  }

  if ( Chan[minor].open == TRUE ) {
      return -EBUSY;
  }

  MOD_INC_USE_COUNT;

  Chan[minor].open = TRUE;              /* The device is open */

  #ifdef DEBUG
      printk("%s: open(): minor %d mode %d.\n", ADAPTER_ID, minor, Chan[minor].mode);
  #endif
  return 0;   
}

/***************************************************************************
 *
 * close() service handler
 *
 ***************************************************************************/

static int adc_close(struct inode *iNode, struct file *filePtr)
{
  int minor = MINOR(iNode->i_rdev);

  #ifdef DEBUG
      printk("%s: close() of minor number %d.\n", ADAPTER_ID, minor);
  #endif

  MOD_DEC_USE_COUNT;
  Chan[minor].open = FALSE;
  return 0;
}

/***************************************************************************
 *
 * iotctl() service handler
 *
 ***************************************************************************/

/* Note:
    Remember that FIOCLEX, FIONCLEX, FIONBIO, and FIOASYN are reserved ioctl cmd numbers
*/

static int adc_ioctl(struct inode *iNode, struct file *filePtr, unsigned int cmd, LONG arg)
{
//  int minor = MINOR(iNode->i_rdev);
#if 0
  switch (cmd) {

    case DAC_XFER:
      /* update all channels at once */
      inb(base_reg);
      break;

    case DIO_SET_MODE:
      if (minor < AD_CHANNELS || minor > (AD_CHANNELS + DIO_PORTS))
        return(-EINVAL);
      arg &= 0x3;

      switch (minor - AD_CHANNELS) {
        case 0:			/* Port A */
          #ifdef DEBUG
          printk("DIO_SET_MODE for Port A\n");
          #endif
          dio_reg &= 0x9f;
          dio_reg |= (arg << 5);
          outb_p(dio_reg, DIO_CNTRL_REG);
          Chan[minor].mode = arg;
          break;

        case 1:			/* Port B */
          #ifdef DEBUG
          printk("DIO_SET_MODE for Port B\n");
          #endif
          if (arg & 0x2)
	    return(-EINVAL);	/* Port B only has Modes 0 & 1 */
          dio_reg &=  0xfb;
          dio_reg |= (arg << 2);
          outb_p(dio_reg, DIO_CNTRL_REG);
          Chan[minor].mode = arg;
          break;

        case 2:			/* Port C */
          #ifdef DEBUG
          printk("DIO_SET_MODE for Port C\n");
          #endif
          if (arg)
	    return(-EINVAL);	/* Port C only has Mode 0 I/O */
          Chan[minor].mode = arg;
          break;

        default:
          #ifdef DEBUG
          printk("DIO_SET_MODE for Invalid Port\n");
          #endif
          return(-EINVAL);	/* Wrong Port Number */
          break;
      }
    break;
    
    case DIO_SET_DIRECTION:
      #ifdef DEBUG
      printk("DIO_SET_DIRECTION: minor = %d, arg = %d\n", (int) minor, (int) arg );
      #endif

      if (minor < AD_CHANNELS || minor > (AD_CHANNELS + DIO_PORTS))
        return(-EINVAL);
    
      switch (minor - AD_CHANNELS) {
        case 0:			/* Port A */
          #ifdef DEBUG
          printk("DIO_SET_DIRECTION for Port A\n");
          #endif
          arg &= 0x1;
          dio_reg &=  0xef;
          dio_reg |= (arg << 4);
          outb_p(dio_reg, DIO_CNTRL_REG);
          break;

        case 1:			/* Port B */
          #ifdef DEBUG
          printk("DIO_SET_DIRECTION for Port B\n");
          #endif
          arg &= 0x1;
          dio_reg &=  0xfd;
          dio_reg |= (arg << 1);
          outb_p(dio_reg, DIO_CNTRL_REG);
          break;

        case 2:			/* Port C */
          #ifdef DEBUG
          printk("DIO_SET_DIRECTION for Port C\n");
          #endif
          switch ( arg ) {
            case PORT_INPUT:         /* CU = IN  CL = IN */
                dio_reg |= 0x9;
                break;
            case PORT_OUTPUT:        /* CU = OUT  CL = OUT */
                dio_reg &= ~(0x9);
                break;
            case LOW_PORT_INPUT:     /* CL = IN */
                dio_reg |= 0x1;
                break;
            case LOW_PORT_OUTPUT:    /* CL = OUT */
                dio_reg &= ~(0x1);
                break;
            case HIGH_PORT_INPUT:    /* CU = IN */
                dio_reg |= 0x8;
                break;
            case HIGH_PORT_OUTPUT:   /* CU = OUT */
                dio_reg &= ~(0x8);
                break;
            default:
                return(-EINVAL);
                break;
          }
          outb_p(dio_reg, DIO_CNTRL_REG);
          break;

        default:
          #ifdef DEBUG
          printk("DIO_SET_DIRECTION for Invalid Port\n");
          #endif
          return(-EINVAL);
          break;
        }
    break;
    
    default:
      return(-EINVAL);
      break;
  }
#endif
  return 0;
}

/***************************************************************************
 *
 *
 ***************************************************************************/

static struct file_operations dac_fops = {
	read:	adc_read,
//	write:	adc_write,
	open:	adc_open,
//	ioctl:	adc_ioctl,
	release:adc_close,
};

#ifdef __cplusplus
extern "C" {
#endif

/***************************************************************************
 *
 * Loads driver. Called when "insmod dda06-16.o" is invoked on the command line.
 *
 ***************************************************************************/

int init_module(void) 
{
  int i,err;

  /* Register as a device with kernel.  */

  if ((err = register_chrdev(MajorNumber, "acl-8216", &dac_fops))) {
      printk("%s: Failure to load module. error %d\n", ADAPTER_ID, -err);
      return err;
  } 

  /* Register the region of iospace with kernel and get base address */

#ifdef ADC_BASE
    base_reg = ADC_BASE;
    if ( check_region(base_reg, PORT_SIZE ) != 0 ) {  /* in use */
        printk("%s: Can't allocate region port address %#x\n", 
                ADAPTER_ID, base_reg );
        return -ENODEV;
    } else {
      request_region( base_reg, PORT_SIZE, "acl-8216" );
    }
#else
  /* Look to see if DAC is installed */
  for ( base_reg = PORT_MIN; base_reg <= PORT_MAX; base_reg += PORT_STEP ) {
      if ( check_region(base_reg, PORT_SIZE ) == 0 ) {  /* not in use */
          request_region(base_reg, PORT_SIZE, "acl-8216");
          if (adc_find( base_reg )) {  /* found the board */
              break;
          } else {
              release_region(base_reg, PORT_SIZE);
          }
      }
  }

  if ( base_reg > PORT_MAX ) {
      printk("%s: No board found from address %#x to %#x.\n", 
             ADAPTER_ID, PORT_MIN, PORT_MAX);
          if (unregister_chrdev(MajorNumber, "acl-8216") != 0) {
              printk("%s: cleanup_module failed\n", ADAPTER_ID);
          }
      return -ENODEV;
  }

#endif

  printk("%s: address=%#x.", ADAPTER_ID, base_reg);
  printk(" 3.4.2000 jakob@hera.ph1.uni-koeln.de\n");

  /* Set all channel structures to show nothing active/open */

  for (i = 0; i < AD_CHANNELS; i++) {
      Chan[i].open = FALSE;
//      Chan[i].addr = base_reg + 2*i;
//      Chan[i].mode = 0;
  }
/*
  Chan[AD_CHANNELS].addr   = DIO_PORTA;
  Chan[AD_CHANNELS+1].addr = DIO_PORTB;
  Chan[AD_CHANNELS+2].addr = DIO_PORTC;

  for ( i = AD_CHANNELS; i < (AD_CHANNELS + DIO_PORTS); i++ ) {
      Chan[i].open = FALSE;
      Chan[i].mode = 0;
      outb_p(0x0, Chan[i].addr);
  }
*/
  return 0;
}

/***************************************************************************
 *
 * Test memory locations for exsistance of dac board.
 *
 ***************************************************************************/
int adc_find( WORD base_reg)
{
	int i;
	signed short r1,r2;

	/* Channel Multiplexer: Channel 0 */
	outb_p( 0x10, base_reg + 10);
	outb_p( 0x01, base_reg + 11); /* Set Software Trigger Mode */
	outb_p( 0xff, base_reg + 12); /* Software Trigger */

	r1=inb_p(base_reg + 8);
	for(i=0;i<10000;i++)
		if( !((r2=inb_p(base_reg + 8))&0x20) ) break;
	if(r1==(r2 + 0x20))
		return TRUE;
	else
		return FALSE;
}

/***************************************************************************
 *
 * Remove driver. Called when "rmmod acl-8216" is run on the command line.
 *
 ***************************************************************************/

void cleanup_module(void) 
{
  if (MOD_IN_USE) {
    printk("%s: device busy, remove delayed\n", ADAPTER_ID);
    return;
  }

  release_region( base_reg, PORT_SIZE );

  if (unregister_chrdev(MajorNumber, "acl-8216") != 0) {
    printk("%s: cleanup_module failed\n", ADAPTER_ID);
  } else {

  #ifdef DEBUG
      printk("%s: module removed.\n", ADAPTER_ID);
  #endif

  }    
}
